# Copyright 2010 Jacob Kaplan-Moss
# Copyright 2011 Nebula, Inc.
# Copyright 2013 Alessio Ababilov
# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""HTTP Exceptions used by keystoneauth1."""

import requests

from keystoneauth1.exceptions import auth
from keystoneauth1.exceptions import base


__all__ = (
    'HttpError',
    'HTTPClientError',
    'BadRequest',
    'Unauthorized',
    'PaymentRequired',
    'Forbidden',
    'NotFound',
    'MethodNotAllowed',
    'NotAcceptable',
    'ProxyAuthenticationRequired',
    'RequestTimeout',
    'Conflict',
    'Gone',
    'LengthRequired',
    'PreconditionFailed',
    'RequestEntityTooLarge',
    'RequestUriTooLong',
    'UnsupportedMediaType',
    'RequestedRangeNotSatisfiable',
    'ExpectationFailed',
    'UnprocessableEntity',
    'TooManyRequests',
    'HttpServerError',
    'InternalServerError',
    'HttpNotImplemented',
    'BadGateway',
    'ServiceUnavailable',
    'GatewayTimeout',
    'HttpVersionNotSupported',
    'from_response',
)


class HttpError(base.ClientException):
    """The base exception class for all HTTP exceptions."""

    http_status = 0
    message = "HTTP Error"

    def __init__(
        self,
        message: str | None = None,
        details: str | None = None,
        response: requests.Response | None = None,
        request_id: str | None = None,
        url: str | None = None,
        method: str | None = None,
        http_status: int | None = None,
        retry_after: int = 0,
    ):
        self.http_status = http_status or self.http_status
        self.message = message or self.message
        self.details = details
        self.request_id = request_id
        self.response = response
        self.url = url
        self.method = method
        formatted_string = f"{self.message} (HTTP {self.http_status})"
        self.retry_after = retry_after
        if request_id:
            formatted_string += f" (Request-ID: {request_id})"
        super().__init__(formatted_string)


class HTTPClientError(HttpError):
    """Client-side HTTP error.

    Exception for cases in which the client seems to have erred.
    """

    message = "HTTP Client Error"


class HttpServerError(HttpError):
    """Server-side HTTP error.

    Exception for cases in which the server is aware that it has
    erred or is incapable of performing the request.
    """

    message = "HTTP Server Error"


class BadRequest(HTTPClientError):
    """HTTP 400 - Bad Request.

    The request cannot be fulfilled due to bad syntax.
    """

    http_status = 400
    message = "Bad Request"


class Unauthorized(HTTPClientError):
    """HTTP 401 - Unauthorized.

    Similar to 403 Forbidden, but specifically for use when authentication
    is required and has failed or has not yet been provided.
    """

    http_status = 401
    message = "Unauthorized"


class PaymentRequired(HTTPClientError):
    """HTTP 402 - Payment Required.

    Reserved for future use.
    """

    http_status = 402
    message = "Payment Required"


class Forbidden(HTTPClientError):
    """HTTP 403 - Forbidden.

    The request was a valid request, but the server is refusing to respond
    to it.
    """

    http_status = 403
    message = "Forbidden"


class NotFound(HTTPClientError):
    """HTTP 404 - Not Found.

    The requested resource could not be found but may be available again
    in the future.
    """

    http_status = 404
    message = "Not Found"


class MethodNotAllowed(HTTPClientError):
    """HTTP 405 - Method Not Allowed.

    A request was made of a resource using a request method not supported
    by that resource.
    """

    http_status = 405
    message = "Method Not Allowed"


class NotAcceptable(HTTPClientError):
    """HTTP 406 - Not Acceptable.

    The requested resource is only capable of generating content not
    acceptable according to the Accept headers sent in the request.
    """

    http_status = 406
    message = "Not Acceptable"


class ProxyAuthenticationRequired(HTTPClientError):
    """HTTP 407 - Proxy Authentication Required.

    The client must first authenticate itself with the proxy.
    """

    http_status = 407
    message = "Proxy Authentication Required"


class RequestTimeout(HTTPClientError):
    """HTTP 408 - Request Timeout.

    The server timed out waiting for the request.
    """

    http_status = 408
    message = "Request Timeout"


class Conflict(HTTPClientError):
    """HTTP 409 - Conflict.

    Indicates that the request could not be processed because of conflict
    in the request, such as an edit conflict.
    """

    http_status = 409
    message = "Conflict"


class Gone(HTTPClientError):
    """HTTP 410 - Gone.

    Indicates that the resource requested is no longer available and will
    not be available again.
    """

    http_status = 410
    message = "Gone"


class LengthRequired(HTTPClientError):
    """HTTP 411 - Length Required.

    The request did not specify the length of its content, which is
    required by the requested resource.
    """

    http_status = 411
    message = "Length Required"


class PreconditionFailed(HTTPClientError):
    """HTTP 412 - Precondition Failed.

    The server does not meet one of the preconditions that the requester
    put on the request.
    """

    http_status = 412
    message = "Precondition Failed"


class RequestEntityTooLarge(HTTPClientError):
    """HTTP 413 - Request Entity Too Large.

    The request is larger than the server is willing or able to process.
    """

    http_status = 413
    message = "Request Entity Too Large"

    def __init__(
        self,
        message: str | None = None,
        details: str | None = None,
        response: requests.Response | None = None,
        request_id: str | None = None,
        url: str | None = None,
        method: str | None = None,
        http_status: int | None = None,
        retry_after: int = 0,
    ):
        try:
            self.retry_after = int(retry_after)
        except (KeyError, ValueError):
            self.retry_after = 0

        super().__init__(
            message=message,
            details=details,
            response=response,
            request_id=request_id,
            url=url,
            method=method,
            http_status=http_status,
        )


class RequestUriTooLong(HTTPClientError):
    """HTTP 414 - Request-URI Too Long.

    The URI provided was too long for the server to process.
    """

    http_status = 414
    message = "Request-URI Too Long"


class UnsupportedMediaType(HTTPClientError):
    """HTTP 415 - Unsupported Media Type.

    The request entity has a media type which the server or resource does
    not support.
    """

    http_status = 415
    message = "Unsupported Media Type"


class RequestedRangeNotSatisfiable(HTTPClientError):
    """HTTP 416 - Requested Range Not Satisfiable.

    The client has asked for a portion of the file, but the server cannot
    supply that portion.
    """

    http_status = 416
    message = "Requested Range Not Satisfiable"


class ExpectationFailed(HTTPClientError):
    """HTTP 417 - Expectation Failed.

    The server cannot meet the requirements of the Expect request-header field.
    """

    http_status = 417
    message = "Expectation Failed"


class UnprocessableEntity(HTTPClientError):
    """HTTP 422 - Unprocessable Entity.

    The request was well-formed but was unable to be followed due to semantic
    errors.
    """

    http_status = 422
    message = "Unprocessable Entity"


class TooManyRequests(HTTPClientError):
    """HTTP 429 - Too Many Requests.

    The user has sent too many requests in a given amount of time.
    The response may include a Retry-After header indicating how long to wait.
    """

    http_status = 429
    message = "Too Many Requests"


class InternalServerError(HttpServerError):
    """HTTP 500 - Internal Server Error.

    A generic error message, given when no more specific message is suitable.
    """

    http_status = 500
    message = "Internal Server Error"


# NotImplemented is a python keyword.
class HttpNotImplemented(HttpServerError):
    """HTTP 501 - Not Implemented.

    The server either does not recognize the request method, or it lacks
    the ability to fulfill the request.
    """

    http_status = 501
    message = "Not Implemented"


class BadGateway(HttpServerError):
    """HTTP 502 - Bad Gateway.

    The server was acting as a gateway or proxy and received an invalid
    response from the upstream server.
    """

    http_status = 502
    message = "Bad Gateway"


class ServiceUnavailable(HttpServerError):
    """HTTP 503 - Service Unavailable.

    The server is currently unavailable.
    """

    http_status = 503
    message = "Service Unavailable"


class GatewayTimeout(HttpServerError):
    """HTTP 504 - Gateway Timeout.

    The server was acting as a gateway or proxy and did not receive a timely
    response from the upstream server.
    """

    http_status = 504
    message = "Gateway Timeout"


class HttpVersionNotSupported(HttpServerError):
    """HTTP 505 - HttpVersion Not Supported.

    The server does not support the HTTP protocol version used in the request.
    """

    http_status = 505
    message = "HTTP Version Not Supported"


# _code_map contains all the classes that have http_status attribute.
_code_map = {
    400: BadRequest,
    401: Unauthorized,
    402: PaymentRequired,
    403: Forbidden,
    404: NotFound,
    405: MethodNotAllowed,
    406: NotAcceptable,
    407: ProxyAuthenticationRequired,
    408: RequestTimeout,
    409: Conflict,
    410: Gone,
    411: LengthRequired,
    412: PreconditionFailed,
    413: RequestEntityTooLarge,
    414: RequestUriTooLong,
    415: UnsupportedMediaType,
    416: RequestedRangeNotSatisfiable,
    417: ExpectationFailed,
    422: UnprocessableEntity,
    429: TooManyRequests,
    500: InternalServerError,
    501: HttpNotImplemented,
    502: BadGateway,
    503: ServiceUnavailable,
    504: GatewayTimeout,
    505: HttpVersionNotSupported,
}


def from_response(
    response: requests.Response, method: str, url: str
) -> HttpError | auth.MissingAuthMethods:
    """Return an instance of :class:`HttpError` or subclass based on response.

    :param response: instance of `requests.Response` class
    :param method: HTTP method used for request
    :param url: URL used for request
    """
    req_id = response.headers.get("x-openstack-request-id")

    message = None
    details = None

    retry_after = 0
    if "retry-after" in response.headers:
        try:
            retry_after = int(response.headers["retry-after"])
        except ValueError:
            pass

    content_type = response.headers.get("Content-Type", "")
    if content_type.startswith("application/json"):
        try:
            body = response.json()
        except ValueError:
            pass
        else:
            if isinstance(body, dict) and isinstance(body.get("error"), dict):
                message = body["error"].get("message")
                details = body["error"].get("details")
            elif isinstance(body, dict) and isinstance(
                body.get("errors"), list
            ):
                # if the error response follows the API SIG guidelines, it
                # will return a list of errors. in this case, only the first
                # error is shown, but if there are multiple the user will be
                # alerted to that fact.
                errors = body["errors"]
                if len(errors) == 0:
                    # just in case we get an empty array
                    message = None
                    details = None
                else:
                    if len(errors) > 1:
                        # if there is more than one error, let the user know
                        # that multiple were seen.
                        msg_hdr = (
                            "Multiple error responses, showing first only: "
                        )
                    else:
                        msg_hdr = ""

                    message = "{}{}".format(msg_hdr, errors[0].get("title"))
                    details = errors[0].get("detail")
            else:
                message = "Unrecognized schema in response body."
    elif content_type.startswith("text/"):
        details = response.text

    # we check explicity for 401 in case of auth receipts
    if (
        response.status_code == 401
        and "Openstack-Auth-Receipt" in response.headers
    ):
        return auth.MissingAuthMethods(response)

    try:
        cls = _code_map[response.status_code]
    except KeyError:
        if 500 <= response.status_code < 600:
            cls = HttpServerError
        elif 400 <= response.status_code < 500:
            cls = HTTPClientError
        else:
            cls = HttpError

    return cls(
        http_status=response.status_code,
        response=response,
        method=method,
        url=url,
        request_id=req_id,
        retry_after=retry_after,
        message=message,
        details=details,
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""A serializer to emit YAML but with request body in nicely formatted JSON."""

import json
import os
import typing as ty

from betamax import cassette
from betamax import serializers
import yaml


def _should_use_block(value: str) -> bool:
    for c in "\u000a\u000d\u001c\u001d\u001e\u0085\u2028\u2029":
        if c in value:
            return True
    return False


def _represent_scalar(
    self: yaml.representer.BaseRepresenter,
    tag: str,
    value: str,
    style: str | None = None,
) -> yaml.representer.ScalarNode:
    if style is None:
        if _should_use_block(value):
            style = '|'
        else:
            style = self.default_style

    node = yaml.representer.ScalarNode(tag, value, style=style)
    if self.alias_key is not None:
        self.represented_objects[self.alias_key] = node
    return node


def _unicode_representer(dumper: yaml.Dumper, uni: str) -> yaml.ScalarNode:
    node = yaml.ScalarNode(tag='tag:yaml.org,2002:str', value=uni)
    return node


def _indent_json(val: str | None) -> str:
    if not val:
        return ''

    return json.dumps(
        json.loads(val),
        indent=2,
        separators=(',', ': '),
        sort_keys=False,
        default=str,
    )


def _is_json_body(interaction: cassette.Interaction) -> bool:
    content_type = interaction['headers'].get('Content-Type', [])
    return 'application/json' in content_type


class YamlJsonSerializer(serializers.BaseSerializer):  # type: ignore
    name = "yamljson"

    @staticmethod
    def generate_cassette_name(
        cassette_library_dir: str, cassette_name: str
    ) -> str:
        return os.path.join(cassette_library_dir, f"{cassette_name}.yaml")

    def serialize(self, cassette_data: dict[str, ty.Any]) -> str:
        # Reserialize internal json with indentation
        for interaction in cassette_data['http_interactions']:
            for key in ('request', 'response'):
                if _is_json_body(interaction[key]):
                    interaction[key]['body']['string'] = _indent_json(
                        interaction[key]['body']['string']
                    )

        class MyDumper(yaml.Dumper):
            """Specialized Dumper which does nice blocks and unicode."""

        yaml.representer.BaseRepresenter.represent_scalar = _represent_scalar  # type: ignore[method-assign]

        MyDumper.add_representer(str, _unicode_representer)

        return yaml.dump(
            cassette_data, Dumper=MyDumper, default_flow_style=False
        )

    def deserialize(self, cassette_data: str) -> ty.Any:
        try:
            deserialized = yaml.safe_load(cassette_data)
        except yaml.error.YAMLError:
            deserialized = None

        if deserialized is not None:
            return deserialized
        return {}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import base64
import copy
import hashlib
import logging
import os
import time
import typing as ty
from urllib import parse as urlparse

import requests

from keystoneauth1 import _utils as utils
from keystoneauth1 import access
from keystoneauth1 import exceptions
from keystoneauth1.identity.v3 import federation
from keystoneauth1 import session as ks_session

LOG = utils.get_logger(__name__)

__all__ = (
    'OidcAuthorizationCode',
    'OidcClientCredentials',
    'OidcPassword',
    'OidcAccessToken',
)

SENSITIVE_KEYS = ("password", "code", "token", "secret")


_OidcBaseT = ty.TypeVar('_OidcBaseT', bound='_OidcBase')


class _OidcBase(federation.FederationBaseAuth, metaclass=abc.ABCMeta):
    """Base class for different OpenID Connect based flows.

    The OpenID Connect specification can be found at::
    ``http://openid.net/specs/openid-connect-core-1_0.html``
    """

    grant_type: ty.ClassVar[str]

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        client_id: str,
        client_secret: str | None,
        access_token_type: str,
        scope: str,
        access_token_endpoint: str | None,
        discovery_endpoint: str | None,
        *,
        trust_id: str | None,
        system_scope: str | None,
        domain_id: str | None,
        domain_name: str | None,
        project_id: str | None,
        project_name: str | None,
        project_domain_id: str | None,
        project_domain_name: str | None,
        reauthenticate: bool,
        include_catalog: bool,
    ):
        """The OpenID Connect plugin expects the following.

        :param auth_url: URL of the Identity Service
        :type auth_url: string

        :param identity_provider: Name of the Identity Provider the client
                                  will authenticate against
        :type identity_provider: string

        :param protocol: Protocol name as configured in keystone
        :type protocol: string

        :param client_id: OAuth 2.0 Client ID
        :type client_id: string

        :param client_secret: OAuth 2.0 Client Secret
        :type client_secret: string

        :param access_token_type: OAuth 2.0 Authorization Server Introspection
                                  token type, it is used to decide which type
                                  of token will be used when processing token
                                  introspection. Valid values are:
                                  "access_token" or "id_token"
        :type access_token_type: string

        :param access_token_endpoint: OpenID Connect Provider Token Endpoint,
                                      for example:
                                      https://localhost:8020/oidc/OP/token
                                      Note that if a discovery document is
                                      provided this value will override
                                      the discovered one.
        :type access_token_endpoint: string

        :param discovery_endpoint: OpenID Connect Discovery Document URL,
            for example:
            https://localhost:8020/oidc/.well-known/openid-configuration
        :type access_token_endpoint: string

        :param scope: OpenID Connect scope that is requested from OP,
                      for example: "openid profile email", defaults to
                      "openid profile". Note that OpenID Connect specification
                      states that "openid" must be always specified.
        :type scope: string
        """
        super().__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            protocol=protocol,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )

        self.client_id = client_id
        self.client_secret = client_secret

        self.discovery_endpoint = discovery_endpoint
        self._discovery_document: dict[str, object] = {}
        self.access_token_endpoint = access_token_endpoint

        self.access_token_type = access_token_type
        self.scope = scope

    def get_unscoped_cache_id_elements(self) -> dict[str, str | None]:
        """Add the OAuth client to what identifies the unscoped token.

        The token is issued to a particular client, so the client id and
        secret join the federation elements from the base. Including the
        secret means rotating it discards any cached token, matching how the
        password plugin treats a changed password. The elements are only
        ever hashed, never stored in the clear.
        """
        elements = super().get_unscoped_cache_id_elements()
        elements['client_id'] = self.client_id
        elements['client_secret'] = self.client_secret
        return elements

    def _get_discovery_document(
        self, session: ks_session.Session
    ) -> dict[str, object]:
        """Get the contents of the OpenID Connect Discovery Document.

        This method grabs the contents of the OpenID Connect Discovery Document
        if a discovery_endpoint was passed to the constructor and returns it as
        a dict, otherwise returns an empty dict. Note that it will fetch the
        discovery document only once, so subsequent calls to this method will
        return the cached result, if any.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a python dictionary containing the discovery document if any,
                  otherwise it will return an empty dict.
        :rtype: dict
        """
        if (
            self.discovery_endpoint is not None
            and not self._discovery_document
        ):
            try:
                resp = session.get(
                    self.discovery_endpoint, authenticated=False
                )
            except exceptions.HttpError:
                LOG.error(
                    "Cannot fetch discovery document %s",
                    self.discovery_endpoint,
                )
                raise

            try:
                self._discovery_document = resp.json()
            except Exception:  # noqa: S110
                # we handle this below
                pass

            if not self._discovery_document:
                raise exceptions.InvalidOidcDiscoveryDocument()

        return self._discovery_document

    def _get_access_token_endpoint(self, session: ks_session.Session) -> str:
        """Get the "token_endpoint" for the OpenID Connect flow.

        This method will return the correct access token endpoint to be used.
        If the user has explicitly passed an access_token_endpoint to the
        constructor that will be returned. If there is no explicit endpoint and
        a discovery url is provided, it will try to get it from the discovery
        document. If nothing is found, an exception will be raised.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :return: the endpoint to use
        :rtype: string
        """
        if self.access_token_endpoint is not None:
            return self.access_token_endpoint

        discovery = self._get_discovery_document(session)
        endpoint = discovery.get("token_endpoint")
        if endpoint is None:
            raise exceptions.OidcAccessTokenEndpointNotFound()
        return ty.cast(str, endpoint)

    def _sanitize(self, data: dict[str, str | None]) -> dict[str, str | None]:
        sanitized = copy.deepcopy(data)
        for key in sanitized:
            if any(s in key for s in SENSITIVE_KEYS):
                sanitized[key] = "***"
        return sanitized

    def _get_access_token(
        self, session: ks_session.Session, payload: dict[str, str | None]
    ) -> str:
        """Exchange a variety of user supplied values for an access token.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :param payload: a dict containing various OpenID Connect values, for
            example::

                {
                    'grant_type': 'password',
                    'username': self.username,
                    'password': self.password,
                    'scope': self.scope,
                }
        :type payload: dict
        """
        if self.client_secret:
            client_auth = (self.client_id, self.client_secret)
        else:
            client_auth = None
            payload.setdefault('client_id', self.client_id)
        access_token_endpoint = self._get_access_token_endpoint(session)

        if LOG.isEnabledFor(logging.DEBUG):
            sanitized_payload = self._sanitize(payload)
            LOG.debug(
                "Making OpenID-Connect authentication request to %s with "
                "data %s",
                access_token_endpoint,
                sanitized_payload,
            )

        op_response = session.post(
            access_token_endpoint,
            requests_auth=client_auth,
            data=payload,
            log=False,
            authenticated=False,
        )
        response = op_response.json()
        if LOG.isEnabledFor(logging.DEBUG):
            sanitized_response = self._sanitize(response)
            LOG.debug(
                "OpenID-Connect authentication response from %s is %s",
                access_token_endpoint,
                sanitized_response,
            )
        access_token = response[self.access_token_type]
        assert isinstance(access_token, str)  # nosec: B101
        return access_token

    def _get_keystone_token(
        self, session: ks_session.Session, access_token: str
    ) -> requests.Response:
        r"""Exchange an access token for a keystone token.

        By Sending the access token in an `Authorization: Bearer` header, to
        an OpenID Connect protected endpoint (Federated Token URL). The
        OpenID Connect server will use the access token to look up information
        about the authenticated user (this technique is called instrospection).
        The output of the instrospection will be an OpenID Connect Claim, that
        will be used against the mapping engine. Should the mapping engine
        succeed, a Keystone token will be presented to the user.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :param access_token: The OpenID Connect access token.
        :type access_token: str
        """
        # use access token against protected URL
        headers = {'Authorization': 'Bearer ' + access_token}
        auth_response = session.post(
            self.federated_token_url, headers=headers, authenticated=False
        )
        return auth_response

    def get_unscoped_auth_ref(
        self, session: ks_session.Session
    ) -> access.AccessInfoV3:
        """Authenticate with OpenID Connect and get back claims.

        This is a multi-step process:

        1.- An access token must be retrieved from the server. In order to do
            so, we need to exchange an authorization grant or refresh token
            with the token endpoint in order to obtain an access token. The
            authorization grant varies from plugin to plugin.

        2.- We then exchange the access token upon accessing the protected
            Keystone endpoint (federated auth URL). This will trigger the
            OpenID Connect Provider to perform a user introspection and
            retrieve information (specified in the scope) about the user in the
            form of an OpenID Connect Claim. These claims will be sent to
            Keystone in the form of environment variables.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a token data representation
        :rtype: :py:class:`keystoneauth1.access.AccessInfoV3`
        """
        # First of all, check if the grant type is supported
        discovery = self._get_discovery_document(session)
        grant_types = ty.cast(
            list[str] | None, discovery.get("grant_types_supported")
        )
        if (
            grant_types
            and self.grant_type is not None
            and self.grant_type not in grant_types
        ):
            raise exceptions.OidcPluginNotSupported()

        # Get the payload
        payload = self.get_payload(session)
        payload.setdefault('grant_type', self.grant_type)

        # get an access token
        access_token = self._get_access_token(session, payload)

        response = self._get_keystone_token(session, access_token)

        # grab the unscoped token
        access_info = access.create(resp=response)
        assert isinstance(access_info, access.AccessInfoV3)  # nosec B101
        return access_info

    @abc.abstractmethod
    def get_payload(
        self, session: ks_session.Session
    ) -> dict[str, str | None]:
        """Get the plugin specific payload for obtainin an access token.

        OpenID Connect supports different grant types. This method should
        prepare the payload that needs to be exchanged with the server in
        order to get an access token for the particular grant type that the
        plugin is implementing.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a python dictionary containing the payload to be exchanged
        :rtype: dict
        """
        raise NotImplementedError()


class OidcPassword(_OidcBase):
    """Implementation for OpenID Connect Resource Owner Password Credential."""

    grant_type = "password"

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        client_id: str,
        client_secret: str,
        access_token_type: str = 'access_token',  # noqa: S107
        scope: str = 'openid profile',
        access_token_endpoint: str | None = None,
        discovery_endpoint: str | None = None,
        username: str | None = None,
        password: str | None = None,
        idp_otp_key: str | None = None,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        """The OpenID Password plugin expects the following.

        :param username: Username used to authenticate
        :type username: string

        :param password: Password used to authenticate
        :type password: string
        """
        super().__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            protocol=protocol,
            client_id=client_id,
            client_secret=client_secret,
            access_token_type=access_token_type,
            scope=scope,
            access_token_endpoint=access_token_endpoint,
            discovery_endpoint=discovery_endpoint,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.username = username
        self.password = password
        self.idp_otp_key = idp_otp_key

    def get_unscoped_cache_id_elements(self) -> dict[str, str | None]:
        """Add the resource owner to what identifies the unscoped token.

        The token stands in for a particular user, so the username and
        password identify it. Without the username two users of the same
        client and identity provider would share one cached unscoped token;
        including the password discards it when the password changes, as the
        password plugin does.
        """
        elements = super().get_unscoped_cache_id_elements()
        elements['username'] = self.username
        elements['password'] = self.password
        return elements

    def get_payload(
        self, session: ks_session.Session
    ) -> dict[str, str | None]:
        """Get an authorization grant for the "password" grant type.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a python dictionary containing the payload to be exchanged
        :rtype: dict
        """
        payload = {
            'username': self.username,
            'password': self.password,
            'scope': self.scope,
        }

        self.manage_otp_from_session_or_request_to_the_user(payload, session)

        return payload

    def manage_otp_from_session_or_request_to_the_user(
        self, payload: dict[str, str | None], session: ks_session.Session
    ) -> None:
        """Get the OTP code from the session or else request to the user.

        When the OS_IDP_OTP_KEY environment variable is set, this method will
        verify if there is an OTP value in the current session, if it exists,
        we use it (the OTP from session) to send to the Identity Provider when
        retrieving the access token. If there is no OTP in the current session,
        we ask the user to enter it (the OTP), and we add it to the session to
        execute the authentication flow.

        The OTP is being stored in the session because in some flows, the CLI
        is doing the authentication process two times, so saving the OTP
        in the session, allow us to use the same OTP in a short time interval,
        avoiding to request it to the user twice in a row.

        :param payload:
        :param session:
        :return:
        """
        if not self.idp_otp_key:
            return

        otp_from_session = getattr(session, 'otp', None)
        if otp_from_session:
            payload[self.idp_otp_key] = otp_from_session
        else:
            payload[self.idp_otp_key] = input(
                "Please, enter the generated OTP code: "
            )
            setattr(session, 'otp', payload[self.idp_otp_key])


class OidcClientCredentials(_OidcBase):
    """Implementation for OpenID Connect Client Credentials."""

    grant_type = 'client_credentials'

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        client_id: str,
        client_secret: str,
        access_token_type: str = 'access_token',  # noqa: S107
        scope: str = 'openid profile',
        access_token_endpoint: str | None = None,
        discovery_endpoint: str | None = None,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        """The OpenID Client Credentials expects the following.

        :param client_id: Client ID used to authenticate
        :type username: string

        :param client_secret: Client Secret used to authenticate
        :type password: string
        """
        super().__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            protocol=protocol,
            client_id=client_id,
            client_secret=client_secret,
            access_token_type=access_token_type,
            scope=scope,
            access_token_endpoint=access_token_endpoint,
            discovery_endpoint=discovery_endpoint,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )

    def get_payload(
        self, session: ks_session.Session
    ) -> dict[str, str | None]:
        """Get an authorization grant for the client credentials grant type.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a python dictionary containing the payload to be exchanged
        :rtype: dict
        """
        payload: dict[str, str | None] = {'scope': self.scope}
        return payload


class OidcAuthorizationCode(_OidcBase):
    """Implementation for OpenID Connect Authorization Code."""

    # The user completes the authorization request in a browser.
    interactive_unscoped_auth = True

    grant_type = 'authorization_code'

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        client_id: str,
        client_secret: str,
        access_token_type: str = 'access_token',  # noqa: S107
        scope: str = 'openid profile',
        access_token_endpoint: str | None = None,
        discovery_endpoint: str | None = None,
        code: str | None = None,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
        redirect_uri: str | None = None,
    ):
        """The OpenID Authorization Code plugin expects the following.

        :param redirect_uri: OpenID Connect Client Redirect URL
        :type redirect_uri: string

        :param code: OAuth 2.0 Authorization Code
        :type code: string

        """
        super().__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            protocol=protocol,
            client_id=client_id,
            client_secret=client_secret,
            access_token_type=access_token_type,
            scope=scope,
            access_token_endpoint=access_token_endpoint,
            discovery_endpoint=discovery_endpoint,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        self.redirect_uri = redirect_uri
        self.code = code

    def get_payload(
        self, session: ks_session.Session
    ) -> dict[str, str | None]:
        """Get an authorization grant for the "authorization_code" grant type.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a python dictionary containing the payload to be exchanged
        :rtype: dict
        """
        payload = {'redirect_uri': self.redirect_uri, 'code': self.code}

        return payload


class OidcAccessToken(_OidcBase):
    """Implementation for OpenID Connect access token reuse."""

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        # client_id and client_id intentionally omitted since they don't make
        # sense with an access token
        access_token_type: str = 'access_token',  # noqa: S107
        scope: str = 'openid profile',
        access_token_endpoint: str | None = None,
        discovery_endpoint: str | None = None,
        access_token: str | None = None,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        """The OpenID Connect plugin based on the Access Token.

        It expects the following:

        :param auth_url: URL of the Identity Service
        :type auth_url: string

        :param identity_provider: Name of the Identity Provider the client
                                  will authenticate against
        :type identity_provider: string

        :param protocol: Protocol name as configured in keystone
        :type protocol: string

        :param access_token: OpenID Connect Access token
        :type access_token: string
        """
        super().__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            protocol=protocol,
            # intentionally set client_id, client_secret to empty since we
            # don't need then
            client_id='',  # nosec B106
            client_secret='',  # nosec B106
            access_token_type=access_token_type,
            scope=scope,
            access_token_endpoint=access_token_endpoint,
            discovery_endpoint=discovery_endpoint,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )
        assert access_token is not None  # nosec B101
        self.access_token = access_token

    def get_unscoped_cache_id_elements(self) -> dict[str, str | None]:
        """Add the access token to what identifies the unscoped token.

        The access token is the whole of the credential here, so it alone
        distinguishes one caller's unscoped token from another's.
        """
        elements = super().get_unscoped_cache_id_elements()
        elements['access_token'] = self.access_token
        return elements

    def get_payload(
        self, session: ks_session.Session
    ) -> dict[str, str | None]:
        """OidcAccessToken does not require a payload."""  # noqa: D403
        return {}

    def get_unscoped_auth_ref(
        self, session: ks_session.Session
    ) -> access.AccessInfoV3:
        """Authenticate with OpenID Connect and get back claims.

        We exchange the access token upon accessing the protected Keystone
        endpoint (federated auth URL). This will trigger the OpenID Connect
        Provider to perform a user introspection and retrieve information
        (specified in the scope) about the user in the form of an OpenID
        Connect Claim. These claims will be sent to Keystone in the form of
        environment variables.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a token data representation
        :rtype: :py:class:`keystoneauth1.access.AccessInfoV3`
        """
        response = self._get_keystone_token(session, self.access_token)
        access_info = access.create(resp=response)
        assert isinstance(access_info, access.AccessInfoV3)  # nosec B101
        return access_info


class OidcDeviceAuthorization(_OidcBase):
    """Implementation for OAuth 2.0 Device Authorization Grant."""

    # The user visits the verification URI and enters a code.
    interactive_unscoped_auth = True

    grant_type = "urn:ietf:params:oauth:grant-type:device_code"
    HEADER_X_FORM = {"Content-Type": "application/x-www-form-urlencoded"}

    def __init__(
        self,
        auth_url: str,
        identity_provider: str,
        protocol: str,
        client_id: str,
        client_secret: str | None = None,
        access_token_type: str = "access_token",  # noqa: S107
        scope: str = 'openid profile',
        access_token_endpoint: str | None = None,
        discovery_endpoint: str | None = None,
        device_authorization_endpoint: str | None = None,
        code_challenge: str | None = None,
        code_challenge_method: str | None = None,
        *,
        trust_id: str | None = None,
        system_scope: str | None = None,
        domain_id: str | None = None,
        domain_name: str | None = None,
        project_id: str | None = None,
        project_name: str | None = None,
        project_domain_id: str | None = None,
        project_domain_name: str | None = None,
        reauthenticate: bool = True,
        include_catalog: bool = True,
    ):
        """The OAuth 2.0 Device Authorization plugin expects the following.

        :param device_authorization_endpoint: OAuth 2.0 Device Authorization
                                  Endpoint, for example:
                                  https://localhost:8020/oidc/authorize/device
                                  Note that if a discovery document is
                                  provided this value will override
                                  the discovered one.
        :type device_authorization_endpoint: string

        :param code_challenge_method: PKCE Challenge Method (RFC 7636).
        :type code_challenge_method: string
        """
        self.device_authorization_endpoint = device_authorization_endpoint
        self.code_challenge_method = code_challenge_method

        super().__init__(
            auth_url=auth_url,
            identity_provider=identity_provider,
            protocol=protocol,
            client_id=client_id,
            client_secret=client_secret,
            access_token_type=access_token_type,
            scope=scope,
            access_token_endpoint=access_token_endpoint,
            discovery_endpoint=discovery_endpoint,
            trust_id=trust_id,
            system_scope=system_scope,
            domain_id=domain_id,
            domain_name=domain_name,
            project_id=project_id,
            project_name=project_name,
            project_domain_id=project_domain_id,
            project_domain_name=project_domain_name,
            reauthenticate=reauthenticate,
            include_catalog=include_catalog,
        )

    def _get_device_authorization_endpoint(
        self, session: ks_session.Session
    ) -> str | None:
        """Get the endpoint for the OAuth 2.0 Device Authorization flow.

        This method will return the correct device authorization endpoint to
        be used.
        If the user has explicitly passed an device_authorization_endpoint to
        the constructor that will be returned. If there is no explicit endpoint
        and a discovery url is provided, it will try to get it from the
        discovery document. If nothing is found, an exception will be raised.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :return: the endpoint to use
        :rtype: string or None if no endpoint is found
        """
        if self.device_authorization_endpoint is not None:
            return self.device_authorization_endpoint

        discovery = self._get_discovery_document(session)
        endpoint = discovery.get("device_authorization_endpoint")
        if endpoint is None:
            raise exceptions.oidc.OidcDeviceAuthorizationEndpointNotFound()
        return ty.cast(str, endpoint)

    def _generate_pkce_verifier(self) -> str:
        """Generate PKCE verifier string as defined in RFC 7636."""
        raw_bytes = 42  # 32 is the minimum from the RFC, let's use a bit more
        _rand = os.urandom(raw_bytes)
        _rand_b64 = base64.urlsafe_b64encode(_rand).decode('ascii')
        code_verifier = _rand_b64.rstrip('=')  # strip padding as RFC says
        return code_verifier

    def _generate_pkce_challenge(self) -> str | None:
        """Generate PKCE challenge string as defined in RFC 7636."""
        if self.code_challenge_method not in ('plain', 'S256'):
            raise exceptions.OidcInvalidCodeChallengeMethod()
        self.code_verifier = self._generate_pkce_verifier()

        if self.code_challenge_method == 'plain':
            return self.code_verifier
        elif self.code_challenge_method == 'S256':
            _tmp = self.code_verifier.encode('ascii')
            _hash = hashlib.sha256(_tmp).digest()
            return base64.urlsafe_b64encode(_hash).decode('ascii').rstrip('=')

        return None

    def get_payload(
        self, session: ks_session.Session
    ) -> dict[str, str | None]:
        """Get an authorization grant for the "device_code" grant type.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :returns: a python dictionary containing the payload to be exchanged
        :rtype: dict
        """
        payload: dict[str, str | None]
        device_authz_endpoint = self._get_device_authorization_endpoint(
            session
        )

        # TODO(stephenfin): This should be an error
        assert device_authz_endpoint is not None  # nosec B101

        if self.client_secret:
            client_auth = (self.client_id, self.client_secret)
        else:
            client_auth = None

        # rfc8628 does not require client_id when a client_secret is provided,
        # but Microsoft EntraID does; these should be ignored by other IDPs
        # when a client secret is provided.
        payload = {'client_id': self.client_id, 'scope': self.scope}

        if self.code_challenge_method:
            self.code_challenge = self._generate_pkce_challenge()
            payload.setdefault(
                'code_challenge_method', self.code_challenge_method
            )
            payload.setdefault('code_challenge', self.code_challenge)
        encoded_payload = urlparse.urlencode(payload)

        if LOG.isEnabledFor(logging.DEBUG):
            sanitized_payload = self._sanitize(payload)
            LOG.debug(
                "Making OpenID-Connect authentication request to %s with "
                "data %s",
                device_authz_endpoint,
                sanitized_payload,
            )
        op_response = session.post(
            device_authz_endpoint,
            requests_auth=client_auth,
            headers=self.HEADER_X_FORM,
            data=encoded_payload,
            log=False,
            authenticated=False,
        )
        if LOG.isEnabledFor(logging.DEBUG):
            sanitized_response = self._sanitize(op_response.json())
            LOG.debug(
                "OpenID-Connect authentication response from %s is %s",
                device_authz_endpoint,
                sanitized_response,
            )

        self.expires_in = int(op_response.json()["expires_in"])
        self.timeout = time.time() + self.expires_in
        self.device_code = op_response.json()["device_code"]
        self.interval = int(op_response.json()["interval"])
        self.user_code = op_response.json()["user_code"]
        self.verification_uri = op_response.json()["verification_uri"]
        self.verification_uri_complete = op_response.json().get(
            "verification_uri_complete"
        )

        payload = {'device_code': self.device_code}
        if self.code_challenge_method:
            payload.setdefault('code_verifier', self.code_verifier)
        return payload

    def _get_access_token(
        self, session: ks_session.Session, payload: dict[str, str | None]
    ) -> str:
        """Poll token endpoint for an access token.

        :param session: a session object to send out HTTP requests.
        :type session: keystoneauth1.session.Session

        :param payload: a dict containing various OpenID Connect values,
            for example::

                {
                    'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
                    'device_code': self.device_code,
                }
        :type payload: dict
        """  # noqa: E501
        # verification_uri_complete is optional and not implemented by EntraID
        if self.verification_uri_complete:
            LOG.warning(
                "To authenticate please go to: %s",
                self.verification_uri_complete,
            )
        else:
            LOG.warning(
                "To authenticate please go to %s and enter the code %s",
                self.verification_uri,
                self.user_code,
            )

        if self.client_secret:
            client_auth = (self.client_id, self.client_secret)
        else:
            client_auth = None

        # rfc8628 does not require client_id when a client_secret is provided,
        # but Microsoft EntraID does; this should be ignored by other IDPs
        # when a client secret is provided.
        payload.setdefault('client_id', self.client_id)

        access_token_endpoint = self._get_access_token_endpoint(session)
        encoded_payload = urlparse.urlencode(payload)

        while time.time() < self.timeout:
            try:
                if LOG.isEnabledFor(logging.DEBUG):
                    sanitized_payload = self._sanitize(payload)
                    LOG.debug(
                        "Making OpenID-Connect authentication request to %s "
                        "with data %s",
                        access_token_endpoint,
                        sanitized_payload,
                    )
                op_response = session.post(
                    access_token_endpoint,
                    requests_auth=client_auth,
                    data=encoded_payload,
                    headers=self.HEADER_X_FORM,
                    log=False,
                    authenticated=False,
                )
                if LOG.isEnabledFor(logging.DEBUG):
                    sanitized_response = self._sanitize(op_response.json())
                    LOG.debug(
                        "OpenID-Connect authentication response from %s is %s",
                        access_token_endpoint,
                        sanitized_response,
                    )
            except exceptions.http.BadRequest as exc:
                if exc.response is None:
                    raise
                error = exc.response.json().get("error")
                if error != "authorization_pending":
                    raise
                time.sleep(self.interval)
                continue
            break
        else:
            if error == "authorization_pending":
                raise exceptions.oidc.OidcDeviceAuthorizationTimeOut()

        access_token = op_response.json()[self.access_token_type]
        assert isinstance(access_token, str)  # nosec B101
        return access_token



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Produce keystone compliant structures for use in testing.

They are part of the public API because they may be relied upon to generate
test tokens for other clients. However they should never be imported into the
main client (keystoneauth or other). Because of this there may be dependencies
from this module on libraries that are only available in testing.
"""

# flake8: noqa: F405

from keystoneauth1.fixture.discovery import *  # noqa
from keystoneauth1.fixture import exception
from keystoneauth1.fixture.plugin import *  # noqa
from keystoneauth1.fixture import v2
from keystoneauth1.fixture import v3


FixtureValidationError = exception.FixtureValidationError
V2Token = v2.Token
V3Token = v3.Token
V3FederationToken = v3.V3FederationToken

__all__ = (
    'DiscoveryList',
    'FixtureValidationError',
    'LoadingFixture',
    'TestPlugin',
    'V2Discovery',
    'V3Discovery',
    'V2Token',
    'V3Token',
    'V3FederationToken',
    'VersionDiscovery',
)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1.exceptions import base


__all__ = (
    'ConnectionError',
    'ConnectTimeout',
    'ConnectFailure',
    'SSLError',
    'RetriableConnectionFailure',
    'UnknownConnectionError',
)


class RetriableConnectionFailure(Exception):
    """A mixin class that implies you can retry the most recent request."""

    pass


class ConnectionError(base.ClientException):
    message = "Cannot connect to API service."


class ConnectTimeout(ConnectionError, RetriableConnectionFailure):
    message = "Timed out connecting to service."


class ConnectFailure(ConnectionError, RetriableConnectionFailure):
    message = "Connection failure that may be retried."


class SSLError(ConnectionError):
    message = "An SSL error occurred."


class UnknownConnectionError(ConnectionError):
    """An error was encountered but we don't know what it is."""

    def __init__(self, msg: str, original: Exception):
        super().__init__(msg)
        self.original = original



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
from unittest import mock
import uuid

import fixtures

from keystoneauth1 import loading
from keystoneauth1.loading import cli
from keystoneauth1.tests.unit.loading import utils


TesterPlugin, TesterLoader = utils.create_plugin(
    opts=[
        loading.Opt(
            'test-opt', help='tester', deprecated=[loading.Opt('test-other')]
        )
    ]
)


class CliTests(utils.TestCase):
    def setUp(self):
        super().setUp()
        self.p = argparse.ArgumentParser()

    def env(self, name, value=None):
        if value is not None:
            # environment variables are always strings
            value = str(value)

        return self.useFixture(fixtures.EnvironmentVariable(name, value))

    def test_creating_with_no_args(self):
        ret = loading.register_auth_argparse_arguments(self.p, [])
        self.assertIsNone(ret)
        self.assertIn('--os-auth-type', self.p.format_usage())

    def test_load_with_nothing(self):
        loading.register_auth_argparse_arguments(self.p, [])
        opts = self.p.parse_args([])
        self.assertIsNone(loading.load_auth_from_argparse_arguments(opts))

    @utils.mock_plugin()
    def test_basic_params_added(self, m):
        name = uuid.uuid4().hex
        argv = ['--os-auth-plugin', name]
        ret = loading.register_auth_argparse_arguments(self.p, argv)
        self.assertIsInstance(ret, utils.MockLoader)

        for n in ('--os-a-int', '--os-a-bool', '--os-a-float'):
            self.assertIn(n, self.p.format_usage())

        m.assert_called_once_with(name)

    @utils.mock_plugin()
    def test_param_loading(self, m):
        name = uuid.uuid4().hex
        argv = [
            '--os-auth-type',
            name,
            '--os-a-int',
            str(self.a_int),
            '--os-a-float',
            str(self.a_float),
            '--os-a-bool',
            str(self.a_bool),
        ]

        klass = loading.register_auth_argparse_arguments(self.p, argv)
        self.assertIsInstance(klass, utils.MockLoader)

        opts = self.p.parse_args(argv)
        self.assertEqual(name, opts.os_auth_type)

        a = loading.load_auth_from_argparse_arguments(opts)
        self.assertTestVals(a)

        self.assertEqual(name, opts.os_auth_type)
        self.assertEqual(str(self.a_int), opts.os_a_int)
        self.assertEqual(str(self.a_float), opts.os_a_float)
        self.assertEqual(str(self.a_bool), opts.os_a_bool)

    @utils.mock_plugin()
    def test_default_options(self, m):
        name = uuid.uuid4().hex
        argv = ['--os-auth-type', name, '--os-a-float', str(self.a_float)]

        klass = loading.register_auth_argparse_arguments(self.p, argv)
        self.assertIsInstance(klass, utils.MockLoader)

        opts = self.p.parse_args(argv)
        self.assertEqual(name, opts.os_auth_type)

        a = loading.load_auth_from_argparse_arguments(opts)

        self.assertEqual(self.a_float, a['a_float'])
        self.assertEqual(3, a['a_int'])

    @utils.mock_plugin()
    def test_with_default_string_value(self, m):
        name = uuid.uuid4().hex
        klass = loading.register_auth_argparse_arguments(
            self.p, [], default=name
        )
        self.assertIsInstance(klass, utils.MockLoader)
        m.assert_called_once_with(name)

    @utils.mock_plugin()
    def test_overrides_default_string_value(self, m):
        name = uuid.uuid4().hex
        default = uuid.uuid4().hex
        argv = ['--os-auth-type', name]
        klass = loading.register_auth_argparse_arguments(
            self.p, argv, default=default
        )
        self.assertIsInstance(klass, utils.MockLoader)
        m.assert_called_once_with(name)

    @utils.mock_plugin()
    def test_with_default_type_value(self, m):
        default = utils.MockLoader()
        klass = loading.register_auth_argparse_arguments(
            self.p, [], default=default
        )
        self.assertIsInstance(klass, utils.MockLoader)
        self.assertEqual(0, m.call_count)

    @utils.mock_plugin()
    def test_overrides_default_type_value(self, m):
        # using this test plugin would fail if called because there
        # is no get_options() function
        class TestLoader:
            pass

        name = uuid.uuid4().hex
        argv = ['--os-auth-type', name]
        klass = loading.register_auth_argparse_arguments(
            self.p, argv, default=TestLoader
        )
        self.assertIsInstance(klass, utils.MockLoader)
        m.assert_called_once_with(name)

    @utils.mock_plugin()
    def test_env_overrides_default_opt(self, m):
        name = uuid.uuid4().hex
        val = uuid.uuid4().hex
        self.env('OS_A_STR', val)

        klass = loading.register_auth_argparse_arguments(
            self.p, [], default=name
        )
        self.assertIsInstance(klass, utils.MockLoader)
        opts = self.p.parse_args([])
        a = loading.load_auth_from_argparse_arguments(opts)

        self.assertEqual(val, a['a_str'])

    def test_deprecated_cli_options(self):
        cli._register_plugin_argparse_arguments(self.p, TesterLoader())
        val = uuid.uuid4().hex
        opts = self.p.parse_args(['--os-test-other', val])
        self.assertEqual(val, opts.os_test_opt)

    def test_deprecated_multi_cli_options(self):
        cli._register_plugin_argparse_arguments(self.p, TesterLoader())
        val1 = uuid.uuid4().hex
        val2 = uuid.uuid4().hex
        # argarse rules say that the last specified wins.
        opts = self.p.parse_args(
            ['--os-test-other', val2, '--os-test-opt', val1]
        )
        self.assertEqual(val1, opts.os_test_opt)

    def test_deprecated_env_options(self):
        val = uuid.uuid4().hex

        with mock.patch.dict('os.environ', {'OS_TEST_OTHER': val}):
            cli._register_plugin_argparse_arguments(self.p, TesterLoader())

        opts = self.p.parse_args([])
        self.assertEqual(val, opts.os_test_opt)

    def test_deprecated_env_multi_options(self):
        val1 = uuid.uuid4().hex
        val2 = uuid.uuid4().hex

        with mock.patch.dict(
            'os.environ', {'OS_TEST_OPT': val1, 'OS_TEST_OTHER': val2}
        ):
            cli._register_plugin_argparse_arguments(self.p, TesterLoader())

        opts = self.p.parse_args([])
        self.assertEqual(val1, opts.os_test_opt)

    def test_adapter_service_type(self):
        argv = ['--os-service-type', 'compute']

        loading.register_adapter_argparse_arguments(self.p, 'compute')

        opts = self.p.parse_args(argv)
        self.assertEqual('compute', opts.os_service_type)
        self.assertFalse(hasattr(opts, 'os_compute_service_type'))

    def test_adapter_service_type_per_service(self):
        argv = ['--os-compute-service-type', 'weirdness']

        loading.register_adapter_argparse_arguments(self.p, 'compute')
        loading.register_service_adapter_argparse_arguments(self.p, 'compute')

        opts = self.p.parse_args(argv)
        self.assertEqual('compute', opts.os_service_type)
        self.assertEqual('weirdness', opts.os_compute_service_type)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import re
from unittest import mock

from testtools import matchers

from keystoneauth1 import adapter
from keystoneauth1 import discover
from keystoneauth1 import exceptions
from keystoneauth1 import fixture
from keystoneauth1 import http_basic
from keystoneauth1 import noauth
from keystoneauth1 import session
from keystoneauth1.tests.unit import utils
from keystoneauth1 import token_endpoint


BASE_HOST = 'http://keystone.example.com'
BASE_URL = f"{BASE_HOST}:5000/"
UPDATED = '2013-03-06T00:00:00Z'

TEST_SERVICE_CATALOG = [
    {
        "endpoints": [
            {
                "adminURL": f"{BASE_HOST}:8774/v1.0",
                "region": "RegionOne",
                "internalURL": f"{BASE_HOST}://127.0.0.1:8774/v1.0",
                "publicURL": f"{BASE_HOST}:8774/v1.0/",
            }
        ],
        "type": "nova_compat",
        "name": "nova_compat",
    },
    {
        "endpoints": [
            {
                "adminURL": "http://nova/novapi/admin",
                "region": "RegionOne",
                "internalURL": "http://nova/novapi/internal",
                "publicURL": "http://nova/novapi/public",
            }
        ],
        "type": "compute",
        "name": "nova",
    },
    {
        "endpoints": [
            {
                "adminURL": "http://glance/glanceapi/admin",
                "region": "RegionOne",
                "internalURL": "http://glance/glanceapi/internal",
                "publicURL": "http://glance/glanceapi/public",
            }
        ],
        "type": "image",
        "name": "glance",
    },
    {
        "endpoints": [
            {
                "adminURL": f"{BASE_HOST}:35357/v2.0",
                "region": "RegionOne",
                "internalURL": f"{BASE_HOST}:5000/v2.0",
                "publicURL": f"{BASE_HOST}:5000/v2.0",
            }
        ],
        "type": "identity",
        "name": "keystone",
    },
    {
        "endpoints": [
            {
                "adminURL": "http://swift/swiftapi/admin",
                "region": "RegionOne",
                "internalURL": "http://swift/swiftapi/internal",
                "publicURL": "http://swift/swiftapi/public",
            }
        ],
        "type": "object-store",
        "name": "swift",
    },
]

V2_URL = f"{BASE_URL}v2.0"
V2_VERSION = fixture.V2Discovery(V2_URL)
V2_VERSION.updated_str = UPDATED

V2_AUTH_RESPONSE = json.dumps(
    {
        "access": {
            "token": {
                "expires": "2020-01-01T00:00:10.000123Z",
                "id": 'fakeToken',
                "tenant": {"id": '1'},
            },
            "user": {"id": 'test'},
            "serviceCatalog": TEST_SERVICE_CATALOG,
        }
    }
)

V3_URL = f"{BASE_URL}v3"
V3_VERSION = fixture.V3Discovery(V3_URL)
V3_MEDIA_TYPES = V3_VERSION.media_types
V3_VERSION.updated_str = UPDATED

V3_AUTH_RESPONSE = json.dumps(
    {
        "token": {
            "methods": ["token", "password"],
            "expires_at": "2020-01-01T00:00:10.000123Z",
            "project": {
                "domain": {"id": '1', "name": 'test-domain'},
                "id": '1',
                "name": 'test-project',
            },
            "user": {
                "domain": {"id": '1', "name": 'test-domain'},
                "id": '1',
                "name": 'test-user',
            },
            "issued_at": "2013-05-29T16:55:21.468960Z",
        }
    }
)

CINDER_EXAMPLES = {
    "versions": [
        {
            "status": discover.Status.CURRENT,
            "updated": "2012-01-04T11:33:21Z",
            "id": "v1.0",
            "links": [{"href": f"{BASE_URL}v1/", "rel": "self"}],
        },
        {
            "status": discover.Status.CURRENT,
            "updated": "2012-11-21T11:33:21Z",
            "id": "v2.0",
            "links": [{"href": f"{BASE_URL}v2/", "rel": "self"}],
        },
        {
            "status": discover.Status.CURRENT,
            "updated": "2012-11-21T11:33:21Z",
            "id": "v3.0",
            "version": "3.27",
            "min_version": "3.0",
            "next_min_version": "3.4",
            "not_before": "2019-12-31",
            "links": [
                {"href": BASE_URL, "rel": "collection"},
                {"href": f"{BASE_URL}v3/", "rel": "self"},
            ],
        },
    ]
}

GLANCE_EXAMPLES = {
    "versions": [
        {
            "status": discover.Status.CURRENT,
            "id": "v2.2",
            "links": [{"href": f"{BASE_URL}v2/", "rel": "self"}],
        },
        {
            "status": discover.Status.SUPPORTED,
            "id": "v2.1",
            "links": [{"href": f"{BASE_URL}v2/", "rel": "self"}],
        },
        {
            "status": discover.Status.SUPPORTED,
            "id": "v2.0",
            "links": [{"href": f"{BASE_URL}v2/", "rel": "self"}],
        },
        {
            "status": discover.Status.CURRENT,
            "id": "v1.1",
            "links": [{"href": f"{BASE_URL}v1/", "rel": "self"}],
        },
        {
            "status": discover.Status.SUPPORTED,
            "id": "v1.0",
            "links": [{"href": f"{BASE_URL}v1/", "rel": "self"}],
        },
    ]
}


def _create_version_list(versions):
    return {'versions': {'values': versions}}


def _create_single_version(version):
    return {'version': version}


V3_VERSION_LIST = _create_version_list([V3_VERSION, V2_VERSION])
V2_VERSION_LIST = _create_version_list([V2_VERSION])

V3_VERSION_ENTRY = _create_single_version(V3_VERSION)
V2_VERSION_ENTRY = _create_single_version(V2_VERSION)


class CatalogHackTests(utils.TestCase):
    TEST_URL = 'http://keystone.server:5000/v2.0'
    OTHER_URL = 'http://other.server:5000/path'

    IDENTITY = 'identity'

    BASE_URL = 'http://keystone.server:5000/'
    V2_URL = BASE_URL + 'v2.0'
    V3_URL = BASE_URL + 'v3'

    def setUp(self):
        super().setUp()
        self.hacks = discover._VersionHacks()
        self.hacks.add_discover_hack(
            self.IDENTITY, re.compile('/v2.0/?$'), '/'
        )

    def test_version_hacks(self):
        self.assertEqual(
            self.BASE_URL,
            self.hacks.get_discover_hack(self.IDENTITY, self.V2_URL),
        )

        self.assertEqual(
            self.BASE_URL,
            self.hacks.get_discover_hack(self.IDENTITY, self.V2_URL + '/'),
        )

        self.assertEqual(
            self.OTHER_URL,
            self.hacks.get_discover_hack(self.IDENTITY, self.OTHER_URL),
        )

    def test_ignored_non_service_type(self):
        self.assertEqual(
            self.V2_URL, self.hacks.get_discover_hack('other', self.V2_URL)
        )


class DiscoverUtils(utils.TestCase):
    def test_version_number(self):
        def assertVersion(out, inp):
            self.assertEqual(out, discover.normalize_version_number(inp))

        def versionRaises(inp):
            self.assertRaises(
                TypeError, discover.normalize_version_number, inp
            )

        assertVersion((1, 2), 'v1.2')
        assertVersion((11, 0), 'v11')
        assertVersion((1, 2), '1.2')
        assertVersion((1, 5, 1), '1.5.1')
        assertVersion((1, 0), '1')
        assertVersion((1, 0), 1)
        assertVersion((5, 2), 5.2)
        assertVersion((3, 20), '3.20')
        assertVersion((6, 1), (6, 1))
        assertVersion((1, 40), [1, 40])
        assertVersion((1, 0), (1,))
        assertVersion((1, 0), ['1'])
        assertVersion((discover.LATEST, discover.LATEST), 'latest')
        assertVersion((discover.LATEST, discover.LATEST), ['latest'])
        assertVersion((discover.LATEST, discover.LATEST), discover.LATEST)
        assertVersion((discover.LATEST, discover.LATEST), (discover.LATEST,))
        assertVersion((10, discover.LATEST), '10.latest')
        assertVersion((10, discover.LATEST), (10, 'latest'))
        assertVersion((10, discover.LATEST), (10, discover.LATEST))

        versionRaises(None)
        versionRaises('hello')
        versionRaises('1.a')
        versionRaises('vacuum')
        versionRaises('')
        versionRaises(('1', 'a'))
        versionRaises([])
        versionRaises([None])

    def test_version_args(self):
        """Validate _normalize_version_args."""

        def assert_min_max(in_ver, in_min, in_max, in_type, out_min, out_max):
            self.assertEqual(
                (out_min, out_max),
                discover._normalize_version_args(
                    in_ver, in_min, in_max, service_type=in_type
                ),
            )

        def normalize_raises(ver, min, max, in_type):
            self.assertRaises(
                ValueError,
                discover._normalize_version_args,
                ver,
                min,
                max,
                service_type=in_type,
            )

        assert_min_max(None, None, None, None, None, None)
        assert_min_max(None, None, 'v1.2', None, None, (1, 2))
        assert_min_max(
            None,
            'v1.2',
            'latest',
            None,
            (1, 2),
            (discover.LATEST, discover.LATEST),
        )
        assert_min_max(None, 'v1.2', '1.6', None, (1, 2), (1, 6))
        assert_min_max(
            None, 'v1.2', '1.latest', None, (1, 2), (1, discover.LATEST)
        )
        assert_min_max(
            None,
            'latest',
            'latest',
            None,
            (discover.LATEST, discover.LATEST),
            (discover.LATEST, discover.LATEST),
        )
        assert_min_max(
            None,
            'latest',
            None,
            None,
            (discover.LATEST, discover.LATEST),
            (discover.LATEST, discover.LATEST),
        )
        assert_min_max(
            None,
            (1, 2),
            None,
            None,
            (1, 2),
            (discover.LATEST, discover.LATEST),
        )
        assert_min_max('', ('1', '2'), (1, 6), None, (1, 2), (1, 6))
        assert_min_max(
            None,
            ('1', '2'),
            (1, discover.LATEST),
            None,
            (1, 2),
            (1, discover.LATEST),
        )
        assert_min_max('v1.2', '', None, None, (1, 2), (1, discover.LATEST))
        assert_min_max(
            '1.latest',
            None,
            '',
            None,
            (1, discover.LATEST),
            (1, discover.LATEST),
        )
        assert_min_max('v1', None, None, None, (1, 0), (1, discover.LATEST))
        assert_min_max(
            'latest',
            None,
            None,
            None,
            (discover.LATEST, discover.LATEST),
            (discover.LATEST, discover.LATEST),
        )
        assert_min_max(
            None, None, 'latest', 'volumev2', (2, 0), (2, discover.LATEST)
        )
        assert_min_max(
            None, None, None, 'volumev2', (2, 0), (2, discover.LATEST)
        )

        normalize_raises('v1', 'v2', None, None)
        normalize_raises('v1', None, 'v2', None)
        normalize_raises(None, 'latest', 'v1', None)
        normalize_raises(None, 'v1.2', 'v1.1', None)
        normalize_raises(None, 'v1.2', 1, None)
        normalize_raises('v2', None, None, 'volumev3')
        normalize_raises('v3', None, None, 'volumev2')
        normalize_raises(None, 'v2', None, 'volumev3')
        normalize_raises(None, None, 'v3', 'volumev2')

    def test_version_to_string(self):
        def assert_string(out, inp):
            self.assertEqual(out, discover.version_to_string(inp))

        assert_string('latest', (discover.LATEST,))
        assert_string('latest', (discover.LATEST, discover.LATEST))
        assert_string(
            'latest', (discover.LATEST, discover.LATEST, discover.LATEST)
        )
        assert_string('1', (1,))
        assert_string('1.2', (1, 2))
        assert_string('1.latest', (1, discover.LATEST))

    def test_version_between(self):
        def good(minver, maxver, cand):
            self.assertTrue(discover.version_between(minver, maxver, cand))

        def bad(minver, maxver, cand):
            self.assertFalse(discover.version_between(minver, maxver, cand))

        def exc(excls, minver, maxver, cand):
            self.assertRaises(
                excls, discover.version_between, minver, maxver, cand
            )

        # candidate required
        exc(ValueError, (1, 0), (1, 0), None)
        exc(ValueError, 'v1.0', '1.0', '')
        # malformed candidate
        exc(TypeError, None, None, 'bogus')
        exc(TypeError, None, None, (1, 'two'))
        # malformed min_version
        exc(TypeError, 'bogus', None, (1, 0))
        exc(TypeError, (1, 'two'), None, (1, 0))
        # malformed max_version
        exc(TypeError, None, 'bogus', (1, 0))
        exc(TypeError, None, (1, 'two'), (1, 0))

        # fail on minimum
        bad((2, 4), None, (1, 55))
        bad('v2.4', '', '2.3')
        bad('latest', None, (2, 3000))
        bad((2, discover.LATEST), '', 'v2.3000')
        bad((2, 3000), '', (1, discover.LATEST))
        bad('latest', None, 'v1000.latest')
        # fail on maximum
        bad(None, (2, 4), (2, 5))
        bad('', 'v2.4', '2.5')
        bad(None, (2, discover.LATEST), (3, 0))
        bad('', '2000.latest', 'latest')

        # candidate matches a bound
        good((1, 0), (1, 0), (1, 0))
        good('1.0', '2.9', '1.0')
        good('v1.0', 'v2.9', 'v2.9')
        # properly in between
        good((1, 0), (1, 10), (1, 2))
        good('1', '2', '1.2')
        # no lower bound
        good(None, (2, 5), (2, 3))
        # no upper bound
        good('2.5', '', '2.6')
        # no bounds at all
        good('', '', 'v1')
        good(None, None, (999, 999))
        good(None, None, 'latest')
        # Various good 'latest' scenarios
        good(
            (discover.LATEST, discover.LATEST),
            (discover.LATEST, discover.LATEST),
            (discover.LATEST, discover.LATEST),
        )
        good(
            (discover.LATEST, discover.LATEST),
            None,
            (discover.LATEST, discover.LATEST),
        )
        good('', 'latest', 'latest')
        good('2.latest', '3.latest', '3.0')
        good('2.latest', None, (55, 66))
        good(None, '3.latest', '3.9999')


class VersionDataTests(utils.TestCase):
    def setUp(self):
        super().setUp()
        self.session = session.Session()

    def test_version_data_basics(self):
        examples = {
            'keystone': V3_VERSION_LIST,
            'cinder': CINDER_EXAMPLES,
            'glance': GLANCE_EXAMPLES,
        }

        for path, data in examples.items():
            url = f"{BASE_URL}{path}"

            mock = self.requests_mock.get(url, status_code=300, json=data)

            disc = discover.Discover(self.session, url)
            raw_data = disc.raw_version_data()
            clean_data = disc.version_data()

            for v in raw_data:
                for n in ('id', 'status', 'links'):
                    msg = f'{n} missing from {path} version data'
                    self.assertThat(
                        v, matchers.Annotate(msg, matchers.Contains(n))
                    )

            for v in clean_data:
                for n in ('version', 'url', 'raw_status'):
                    msg = f'{n} missing from {path} version data'
                    self.assertThat(
                        v, matchers.Annotate(msg, matchers.Contains(n))
                    )

            self.assertTrue(mock.called_once)

    def test_version_data_override_version_url(self):
        # if the request url is versioned already, just return it.
        self.requests_mock.get(
            V3_URL,
            status_code=200,
            json={
                'version': fixture.V3Discovery('http://override/identity/v3')
            },
        )

        disc = discover.Discover(self.session, V3_URL)
        version_data = disc.version_data()

        for v in version_data:
            self.assertEqual(v['version'], (3, 0))
            self.assertEqual(v['status'], discover.Status.CURRENT)
            self.assertEqual(v['raw_status'], 'stable')
            self.assertEqual(v['url'], V3_URL)

        # if the request url is not versioned, just add version info to it.(
        # do not changed the url's netloc or path)
        self.requests_mock.get(
            BASE_URL,
            status_code=200,
            json={
                'version': fixture.V3Discovery('http://override/identity/v3')
            },
        )

        disc = discover.Discover(self.session, BASE_URL)
        version_data = disc.version_data()

        for v in version_data:
            self.assertEqual(v['version'], (3, 0))
            self.assertEqual(v['status'], discover.Status.CURRENT)
            self.assertEqual(v['raw_status'], 'stable')
            self.assertEqual(v['url'], V3_URL)

    def test_version_data_unknown(self):
        discovery_fixture = fixture.V3Discovery(V3_URL)
        discovery_fixture.status = 'hungry'
        discovery_doc = _create_single_version(discovery_fixture)

        self.requests_mock.get(V3_URL, status_code=200, json=discovery_doc)

        disc = discover.Discover(self.session, V3_URL)
        clean_data = disc.version_data(allow_unknown=True)

        self.assertEqual(discover.Status.UNKNOWN, clean_data[0]['status'])

    def test_version_data_individual(self):
        mock = self.requests_mock.get(
            V3_URL, status_code=200, json=V3_VERSION_ENTRY
        )

        disc = discover.Discover(self.session, V3_URL)
        raw_data = disc.raw_version_data()
        clean_data = disc.version_data()

        for v in raw_data:
            self.assertEqual(v['id'], 'v3.0')
            self.assertEqual(v['status'], 'stable')
            self.assertIn('media-types', v)
            self.assertIn('links', v)

        for v in clean_data:
            self.assertEqual(v['version'], (3, 0))
            self.assertEqual(v['status'], discover.Status.CURRENT)
            self.assertEqual(v['raw_status'], 'stable')
            self.assertEqual(v['url'], V3_URL)

        self.assertTrue(mock.called_once)

    def test_version_data_legacy_ironic_no_override(self):
        """Validate detection of legacy Ironic microversion ranges."""
        ironic_url = 'https://bare-metal.example.com/v1/'
        self.requests_mock.get(
            ironic_url,
            status_code=200,
            json={'id': 'v1', 'links': [{"href": ironic_url, "rel": "self"}]},
            headers={
                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
            },
        )

        plugin = noauth.NoAuth()
        a = adapter.Adapter(
            self.session, auth=plugin, service_type='baremetal'
        )

        self.assertIsNone(a.get_api_major_version())

    def test_version_data_ironic_microversions(self):
        """Validate detection of Ironic microversion ranges."""
        ironic_url = 'https://bare-metal.example.com/v1/'
        self.requests_mock.get(
            ironic_url,
            status_code=200,
            json={
                'id': 'v1',
                'version': {
                    'id': 'v1',
                    'links': [{"href": ironic_url, "rel": "self"}],
                    'version': '1.40',
                    'min_version': '1.10',
                    'status': 'CURRENT',
                },
                'links': [{"href": ironic_url, "rel": "self"}],
            },
            # Keep headers so we can verify that body trumps headers
            headers={
                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
            },
        )

        self.assertEqual(
            [
                {
                    'collection': None,
                    'version': (1, 0),
                    'url': ironic_url,
                    'status': discover.Status.CURRENT,
                    'raw_status': discover.Status.CURRENT,
                    'min_microversion': (1, 10),
                    'max_microversion': (1, 40),
                    'next_min_version': None,
                    'not_before': None,
                }
            ],
            discover.Discover(self.session, ironic_url).version_data(),
        )

    def test_version_data_legacy_ironic_microversions(self):
        """Validate detection of legacy Ironic microversion ranges."""
        ironic_url = 'https://bare-metal.example.com/v1/'
        self.requests_mock.get(
            ironic_url,
            status_code=200,
            json={'id': 'v1', 'links': [{"href": ironic_url, "rel": "self"}]},
            headers={
                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
            },
        )

        self.assertEqual(
            [
                {
                    'collection': None,
                    'version': (1, 0),
                    'url': ironic_url,
                    'status': discover.Status.CURRENT,
                    'raw_status': discover.Status.CURRENT,
                    'min_microversion': (1, 3),
                    'max_microversion': (1, 21),
                    'next_min_version': None,
                    'not_before': None,
                }
            ],
            discover.Discover(self.session, ironic_url).version_data(),
        )

    def test_version_data_microversions(self):
        """Validate [min_|max_]version conversion to {min|max}_microversion."""

        def setup_mock(versions_in):
            # Set up the test data with the input version data
            jsondata = {
                "versions": [
                    dict(
                        {
                            "status": discover.Status.CURRENT,
                            "id": "v2.2",
                            "links": [{"href": V3_URL, "rel": "self"}],
                        },
                        **versions_in,
                    )
                ]
            }
            self.requests_mock.get(V3_URL, status_code=200, json=jsondata)

        def test_ok(versions_in, versions_out):
            setup_mock(versions_in)
            # Ensure the output contains the expected microversions
            self.assertEqual(
                [
                    dict(
                        {
                            'collection': None,
                            'version': (2, 2),
                            'url': V3_URL,
                            'status': discover.Status.CURRENT,
                            'raw_status': discover.Status.CURRENT,
                        },
                        **versions_out,
                    )
                ],
                discover.Discover(self.session, V3_URL).version_data(),
            )

        def test_exc(versions_in):
            setup_mock(versions_in)
            # Ensure TypeError is raised
            self.assertRaises(
                TypeError, discover.Discover(self.session, V3_URL).version_data
            )

        # no version info in input
        test_ok(
            {},
            {
                'min_microversion': None,
                'max_microversion': None,
                'next_min_version': None,
                'not_before': None,
            },
        )

        # version => max_microversion
        test_ok(
            {'version': '2.2'},
            {
                'min_microversion': None,
                'max_microversion': (2, 2),
                'next_min_version': None,
                'not_before': None,
            },
        )

        # max_version supersedes version (even if malformed).  min_version &
        # normalization.
        test_ok(
            {'min_version': '2', 'version': 'foo', 'max_version': '2.2'},
            {
                'min_microversion': (2, 0),
                'max_microversion': (2, 2),
                'next_min_version': None,
                'not_before': None,
            },
        )

        # Edge case: min/max_version ignored if present but "empty"; version
        # used for max_microversion.
        test_ok(
            {'min_version': '', 'version': '2.1', 'max_version': ''},
            {
                'min_microversion': None,
                'max_microversion': (2, 1),
                'next_min_version': None,
                'not_before': None,
            },
        )

        # next_min_version set
        test_ok(
            {
                'min_version': '2',
                'max_version': '2.2',
                'next_min_version': '2.1',
                'not_before': '2019-07-01',
            },
            {
                'min_microversion': (2, 0),
                'max_microversion': (2, 2),
                'next_min_version': (2, 1),
                'not_before': '2019-07-01',
            },
        )

        # Badly-formatted min_version
        test_exc({'min_version': 'foo', 'max_version': '2.1'})

        # Badly-formatted max_version
        test_exc({'min_version': '2.1', 'max_version': 'foo'})

        # Badly-formatted version (when max_version omitted)
        test_exc({'min_version': '2.1', 'version': 'foo'})

        # Badly-formatted next_min_version
        test_exc({'next_min_version': 'bogus', 'not_before': '2019-07-01'})

    def test_endpoint_data_noauth_discover(self):
        mock = self.requests_mock.get(
            BASE_URL, status_code=200, json=V3_VERSION_LIST
        )
        self.requests_mock.get(V3_URL, status_code=200, json=V3_VERSION_ENTRY)

        plugin = noauth.NoAuth(endpoint=BASE_URL)
        data = plugin.get_endpoint_data(self.session)

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
        self.assertEqual(plugin.get_endpoint(self.session), BASE_URL)

        self.assertTrue(mock.called_once)

    def test_endpoint_data_noauth_versioned_discover(self):
        self.requests_mock.get(BASE_URL, status_code=200, json=V3_VERSION_LIST)
        self.requests_mock.get(V3_URL, status_code=200, json=V3_VERSION_ENTRY)

        plugin = noauth.NoAuth(endpoint=V3_URL)
        data = plugin.get_endpoint_data(self.session)

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)

    def test_endpoint_data_noauth_no_discover(self):
        plugin = noauth.NoAuth(endpoint=V3_URL)
        data = plugin.get_endpoint_data(self.session, discover_versions=False)

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)

    def test_endpoint_data_noauth_override_no_discover(self):
        plugin = noauth.NoAuth()
        data = plugin.get_endpoint_data(
            self.session, endpoint_override=V3_URL, discover_versions=False
        )

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(
            plugin.get_endpoint(self.session, endpoint_override=V3_URL), V3_URL
        )

    def test_endpoint_data_http_basic_discover(self):
        self.requests_mock.get(BASE_URL, status_code=200, json=V3_VERSION_LIST)
        self.requests_mock.get(V3_URL, status_code=200, json=V3_VERSION_ENTRY)

        plugin = http_basic.HTTPBasicAuth(endpoint=V3_URL)
        data = plugin.get_endpoint_data(self.session)

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)

    def test_endpoint_data_http_basic_no_discover(self):
        plugin = http_basic.HTTPBasicAuth(endpoint=V3_URL)
        data = plugin.get_endpoint_data(self.session, discover_versions=False)

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)

    def test_endpoint_data_http_basic_override_no_discover(self):
        plugin = http_basic.HTTPBasicAuth()
        data = plugin.get_endpoint_data(
            self.session, endpoint_override=V3_URL, discover_versions=False
        )

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(
            plugin.get_api_major_version(
                self.session, endpoint_override=V3_URL
            ),
            (3, 0),
        )
        self.assertEqual(
            plugin.get_endpoint(self.session, endpoint_override=V3_URL), V3_URL
        )

    def test_endpoint_data_noauth_adapter(self):
        self.requests_mock.get(BASE_URL, status_code=200, json=V3_VERSION_LIST)
        self.requests_mock.get(V3_URL, status_code=200, json=V3_VERSION_ENTRY)

        client = adapter.Adapter(
            session.Session(noauth.NoAuth()), endpoint_override=BASE_URL
        )
        data = client.get_endpoint_data()

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(client.get_api_major_version(), (3, 0))
        self.assertEqual(client.get_endpoint(), BASE_URL)

    def test_endpoint_data_noauth_versioned_adapter(self):
        mock = self.requests_mock.get(
            V3_URL, status_code=200, json=V3_VERSION_ENTRY
        )

        client = adapter.Adapter(
            session.Session(noauth.NoAuth()), endpoint_override=V3_URL
        )
        data = client.get_endpoint_data()

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(client.get_api_major_version(), (3, 0))
        self.assertEqual(client.get_endpoint(), V3_URL)

        self.assertTrue(mock.called_once)

    def test_endpoint_data_token_endpoint_discover(self):
        mock = self.requests_mock.get(
            V3_URL, status_code=200, json=V3_VERSION_ENTRY
        )
        plugin = token_endpoint.Token(endpoint=V3_URL, token='bogus')
        data = plugin.get_endpoint_data(self.session)

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)

        self.assertTrue(mock.called_once)

    def test_endpoint_data_token_endpoint_no_discover(self):
        plugin = token_endpoint.Token(endpoint=V3_URL, token='bogus')
        data = plugin.get_endpoint_data(self.session, discover_versions=False)

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)

    def test_endpoint_data_token_endpoint_adapter(self):
        mock = self.requests_mock.get(
            V3_URL, status_code=200, json=V3_VERSION_ENTRY
        )
        plugin = token_endpoint.Token(endpoint=V3_URL, token='bogus')

        client = adapter.Adapter(session.Session(plugin))
        data = client.get_endpoint_data()

        self.assertEqual(data.api_version, (3, 0))
        self.assertEqual(data.url, V3_URL)
        self.assertEqual(client.get_api_major_version(), (3, 0))
        self.assertEqual(client.get_endpoint(), V3_URL)

        self.assertTrue(mock.called_once)

    def test_data_for_url(self):
        mock = self.requests_mock.get(
            V3_URL, status_code=200, json=V3_VERSION_ENTRY
        )

        disc = discover.Discover(self.session, V3_URL)
        for url in (V3_URL, V3_URL + '/'):
            data = disc.versioned_data_for(url=url)
            self.assertEqual(data['version'], (3, 0))
            self.assertEqual(data['raw_status'], 'stable')
            self.assertEqual(data['url'], V3_URL)

        self.assertTrue(mock.called_once)

    def test_data_for_no_version(self):
        mock = self.requests_mock.get(
            V3_URL, status_code=200, json=V3_VERSION_ENTRY
        )

        disc = discover.Discover(self.session, V3_URL)

        data = disc.versioned_data_for()
        self.assertEqual(data['version'], (3, 0))
        self.assertEqual(data['raw_status'], 'stable')
        self.assertEqual(data['url'], V3_URL)
        self.assertRaises(TypeError, disc.data_for, version=None)

        self.assertTrue(mock.called_once)

    def test_keystone_version_data(self):
        mock = self.requests_mock.get(
            BASE_URL, status_code=300, json=V3_VERSION_LIST
        )

        disc = discover.Discover(self.session, BASE_URL)
        raw_data = disc.raw_version_data()
        clean_data = disc.version_data()

        self.assertEqual(2, len(raw_data))
        self.assertEqual(2, len(clean_data))

        for v in raw_data:
            self.assertIn(v['id'], ('v2.0', 'v3.0'))
            self.assertEqual(v['updated'], UPDATED)
            self.assertEqual(v['status'], 'stable')

            if v['id'] == 'v3.0':
                self.assertEqual(v['media-types'], V3_MEDIA_TYPES)

        for v in clean_data:
            self.assertIn(v['version'], ((2, 0), (3, 0)))
            self.assertEqual(v['raw_status'], 'stable')

        valid_v3_versions = (
            disc.data_for('v3.0'),
            disc.data_for('3.latest'),
            disc.data_for('latest'),
            disc.versioned_data_for(
                min_version='v3.0', max_version='v3.latest'
            ),
            disc.versioned_data_for(min_version='3'),
            disc.versioned_data_for(min_version='3.latest'),
            disc.versioned_data_for(min_version='latest'),
            disc.versioned_data_for(
                min_version='3.latest', max_version='latest'
            ),
            disc.versioned_data_for(
                min_version='latest', max_version='latest'
            ),
            disc.versioned_data_for(min_version=2),
            disc.versioned_data_for(min_version='2.latest'),
        )
        for version in valid_v3_versions:
            self.assertEqual((3, 0), version['version'])
            self.assertEqual('stable', version['raw_status'])
            self.assertEqual(V3_URL, version['url'])

        valid_v2_versions = (
            disc.data_for(2),
            disc.data_for('2.latest'),
            disc.versioned_data_for(
                min_version=2, max_version=(2, discover.LATEST)
            ),
            disc.versioned_data_for(
                min_version='2.latest', max_version='2.latest'
            ),
        )
        for version in valid_v2_versions:
            self.assertEqual((2, 0), version['version'])
            self.assertEqual('stable', version['raw_status'])
            self.assertEqual(V2_URL, version['url'])

        self.assertIsNone(disc.url_for('v4'))
        self.assertIsNone(
            disc.versioned_url_for(min_version='v4', max_version='v4.latest')
        )
        self.assertEqual(V3_URL, disc.url_for('v3'))
        self.assertEqual(
            V3_URL,
            disc.versioned_url_for(min_version='v3', max_version='v3.latest'),
        )
        self.assertEqual(V2_URL, disc.url_for('v2'))
        self.assertEqual(
            V2_URL,
            disc.versioned_url_for(min_version='v2', max_version='v2.latest'),
        )

        self.assertTrue(mock.called_once)

    def test_cinder_version_data(self):
        mock = self.requests_mock.get(
            BASE_URL, status_code=300, json=CINDER_EXAMPLES
        )

        disc = discover.Discover(self.session, BASE_URL)
        raw_data = disc.raw_version_data()
        clean_data = disc.version_data()

        self.assertEqual(3, len(raw_data))

        for v in raw_data:
            self.assertEqual(v['status'], discover.Status.CURRENT)
            if v['id'] == 'v1.0':
                self.assertEqual(v['updated'], '2012-01-04T11:33:21Z')
            elif v['id'] == 'v2.0':
                self.assertEqual(v['updated'], '2012-11-21T11:33:21Z')
            elif v['id'] == 'v3.0':
                self.assertEqual(v['updated'], '2012-11-21T11:33:21Z')
            else:
                self.fail("Invalid version found")

        v1_url = f"{BASE_URL}v1/"
        v2_url = f"{BASE_URL}v2/"
        v3_url = f"{BASE_URL}v3/"

        self.assertEqual(
            clean_data,
            [
                {
                    'collection': None,
                    'max_microversion': None,
                    'min_microversion': None,
                    'next_min_version': None,
                    'not_before': None,
                    'version': (1, 0),
                    'url': v1_url,
                    'status': discover.Status.CURRENT,
                    'raw_status': discover.Status.CURRENT,
                },
                {
                    'collection': None,
                    'max_microversion': None,
                    'min_microversion': None,
                    'next_min_version': None,
                    'not_before': None,
                    'version': (2, 0),
                    'url': v2_url,
                    'status': discover.Status.CURRENT,
                    'raw_status': discover.Status.CURRENT,
                },
                {
                    'collection': BASE_URL,
                    'max_microversion': (3, 27),
                    'min_microversion': (3, 0),
                    'next_min_version': (3, 4),
                    'not_before': '2019-12-31',
                    'version': (3, 0),
                    'url': v3_url,
                    'status': discover.Status.CURRENT,
                    'raw_status': discover.Status.CURRENT,
                },
            ],
        )

        for version in (
            disc.data_for('v2.0'),
            disc.versioned_data_for(
                min_version='v2.0', max_version='v2.latest'
            ),
        ):
            self.assertEqual((2, 0), version['version'])
            self.assertEqual(discover.Status.CURRENT, version['raw_status'])
            self.assertEqual(v2_url, version['url'])

        for version in (
            disc.data_for(1),
            disc.versioned_data_for(
                min_version=(1,), max_version=(1, discover.LATEST)
            ),
        ):
            self.assertEqual((1, 0), version['version'])
            self.assertEqual(discover.Status.CURRENT, version['raw_status'])
            self.assertEqual(v1_url, version['url'])

        self.assertIsNone(disc.url_for('v4'))
        self.assertIsNone(
            disc.versioned_url_for(min_version='v4', max_version='v4.latest')
        )
        self.assertEqual(v3_url, disc.url_for('v3'))
        self.assertEqual(
            v3_url,
            disc.versioned_url_for(min_version='v3', max_version='v3.latest'),
        )
        self.assertEqual(v2_url, disc.url_for('v2'))
        self.assertEqual(
            v2_url,
            disc.versioned_url_for(min_version='v2', max_version='v2.latest'),
        )
        self.assertEqual(v1_url, disc.url_for('v1'))
        self.assertEqual(
            v1_url,
            disc.versioned_url_for(min_version='v1', max_version='v1.latest'),
        )

        self.assertTrue(mock.called_once)

    def test_glance_version_data(self):
        mock = self.requests_mock.get(
            BASE_URL, status_code=200, json=GLANCE_EXAMPLES
        )

        disc = discover.Discover(self.session, BASE_URL)
        raw_data = disc.raw_version_data()
        clean_data = disc.version_data()

        self.assertEqual(5, len(raw_data))

        for v in raw_data:
            if v['id'] in ('v2.2', 'v1.1'):
                self.assertEqual(v['status'], discover.Status.CURRENT)
            elif v['id'] in ('v2.1', 'v2.0', 'v1.0'):
                self.assertEqual(v['status'], discover.Status.SUPPORTED)
            else:
                self.fail("Invalid version found")

        v1_url = f'{BASE_URL}v1/'
        v2_url = f'{BASE_URL}v2/'

        self.assertEqual(
            clean_data,
            [
                {
                    'collection': None,
                    'max_microversion': None,
                    'min_microversion': None,
                    'next_min_version': None,
                    'not_before': None,
                    'version': (1, 0),
                    'url': v1_url,
                    'status': discover.Status.SUPPORTED,
                    'raw_status': discover.Status.SUPPORTED,
                },
                {
                    'collection': None,
                    'max_microversion': None,
                    'min_microversion': None,
                    'next_min_version': None,
                    'not_before': None,
                    'version': (1, 1),
                    'url': v1_url,
                    'status': discover.Status.CURRENT,
                    'raw_status': discover.Status.CURRENT,
                },
                {
                    'collection': None,
                    'max_microversion': None,
                    'min_microversion': None,
                    'next_min_version': None,
                    'not_before': None,
                    'version': (2, 0),
                    'url': v2_url,
                    'status': discover.Status.SUPPORTED,
                    'raw_status': discover.Status.SUPPORTED,
                },
                {
                    'collection': None,
                    'max_microversion': None,
                    'min_microversion': None,
                    'next_min_version': None,
                    'not_before': None,
                    'version': (2, 1),
                    'url': v2_url,
                    'status': discover.Status.SUPPORTED,
                    'raw_status': discover.Status.SUPPORTED,
                },
                {
                    'collection': None,
                    'max_microversion': None,
                    'min_microversion': None,
                    'next_min_version': None,
                    'not_before': None,
                    'version': (2, 2),
                    'url': v2_url,
                    'status': discover.Status.CURRENT,
                    'raw_status': discover.Status.CURRENT,
                },
            ],
        )

        for ver in (2, 2.1, 2.2):
            for version in (
                disc.data_for(ver),
                disc.versioned_data_for(
                    min_version=ver, max_version=(2, discover.LATEST)
                ),
            ):
                self.assertEqual((2, 2), version['version'])
                self.assertEqual(
                    discover.Status.CURRENT, version['raw_status']
                )
                self.assertEqual(v2_url, version['url'])
                self.assertEqual(v2_url, disc.url_for(ver))

        for ver in (1, 1.1):
            for version in (
                disc.data_for(ver),
                disc.versioned_data_for(
                    min_version=ver, max_version=(1, discover.LATEST)
                ),
            ):
                self.assertEqual((1, 1), version['version'])
                self.assertEqual(
                    discover.Status.CURRENT, version['raw_status']
                )
                self.assertEqual(v1_url, version['url'])
                self.assertEqual(v1_url, disc.url_for(ver))

        self.assertIsNone(disc.url_for('v3'))
        self.assertIsNone(
            disc.versioned_url_for(min_version='v3', max_version='v3.latest')
        )
        self.assertIsNone(disc.url_for('v2.3'))
        self.assertIsNone(
            disc.versioned_url_for(min_version='v2.3', max_version='v2.latest')
        )

        self.assertTrue(mock.called_once)

    def test_allow_deprecated(self):
        status = 'deprecated'
        version_list = [
            {
                'id': 'v3.0',
                'links': [{'href': V3_URL, 'rel': 'self'}],
                'media-types': V3_MEDIA_TYPES,
                'status': status,
                'updated': UPDATED,
            }
        ]
        self.requests_mock.get(BASE_URL, json={'versions': version_list})

        disc = discover.Discover(self.session, BASE_URL)

        # deprecated is allowed by default
        versions = disc.version_data(allow_deprecated=False)
        self.assertEqual(0, len(versions))

        versions = disc.version_data(allow_deprecated=True)
        self.assertEqual(1, len(versions))
        self.assertEqual(status, versions[0]['raw_status'])
        self.assertEqual(V3_URL, versions[0]['url'])
        self.assertEqual((3, 0), versions[0]['version'])

    def test_allow_experimental(self):
        status = 'experimental'
        version_list = [
            {
                'id': 'v3.0',
                'links': [{'href': V3_URL, 'rel': 'self'}],
                'media-types': V3_MEDIA_TYPES,
                'status': status,
                'updated': UPDATED,
            }
        ]
        self.requests_mock.get(BASE_URL, json={'versions': version_list})

        disc = discover.Discover(self.session, BASE_URL)

        versions = disc.version_data()
        self.assertEqual(0, len(versions))

        versions = disc.version_data(allow_experimental=True)
        self.assertEqual(1, len(versions))
        self.assertEqual(status, versions[0]['raw_status'])
        self.assertEqual(V3_URL, versions[0]['url'])
        self.assertEqual((3, 0), versions[0]['version'])

    def test_allow_unknown(self):
        status = 'abcdef'
        version_list = fixture.DiscoveryList(
            BASE_URL, v2=False, v3_status=status
        )
        self.requests_mock.get(BASE_URL, json=version_list)

        disc = discover.Discover(self.session, BASE_URL)

        versions = disc.version_data()
        self.assertEqual(0, len(versions))

        versions = disc.version_data(allow_unknown=True)
        self.assertEqual(1, len(versions))
        self.assertEqual(status, versions[0]['raw_status'])
        self.assertEqual(V3_URL, versions[0]['url'])
        self.assertEqual((3, 0), versions[0]['version'])

    def test_ignoring_invalid_links(self):
        version_list = [
            {
                'id': 'v3.0',
                'links': [{'href': V3_URL, 'rel': 'self'}],
                'media-types': V3_MEDIA_TYPES,
                'status': 'stable',
                'updated': UPDATED,
            },
            {
                'id': 'v3.1',
                'media-types': V3_MEDIA_TYPES,
                'status': 'stable',
                'updated': UPDATED,
            },
            {
                'media-types': V3_MEDIA_TYPES,
                'status': 'stable',
                'updated': UPDATED,
                'links': [{'href': V3_URL, 'rel': 'self'}],
            },
        ]

        self.requests_mock.get(BASE_URL, json={'versions': version_list})

        disc = discover.Discover(self.session, BASE_URL)

        # raw_version_data will return all choices, even invalid ones
        versions = disc.raw_version_data()
        self.assertEqual(3, len(versions))

        # only the version with both id and links will be actually returned
        versions = disc.version_data()
        self.assertEqual(1, len(versions))


class EndpointDataTests(utils.TestCase):
    def setUp(self):
        super().setUp()
        self.session = session.Session()

    @mock.patch('keystoneauth1.discover.get_discovery')
    @mock.patch(
        'keystoneauth1.discover.EndpointData._get_discovery_url_choices'
    )
    def test_run_discovery_cache(self, mock_url_choices, mock_get_disc):
        # get_discovery raises so we keep looping
        mock_get_disc.side_effect = exceptions.DiscoveryFailure()
        # Duplicate 'url1' in here to validate the cache behavior
        mock_url_choices.return_value = ('url1', 'url2', 'url1', 'url3')
        epd = discover.EndpointData()
        epd._run_discovery(
            session='sess',
            cache='cache',
            min_version='min',
            max_version='max',
            project_id='projid',
            allow_version_hack='allow_hack',
            discover_versions='disc_vers',
        )
        # Only one call with 'url1'
        self.assertEqual(3, mock_get_disc.call_count)
        mock_get_disc.assert_has_calls(
            [
                mock.call('sess', url, cache='cache', authenticated=False)
                for url in ('url1', 'url2', 'url3')
            ]
        )

    def test_run_discovery_auth(self):
        url = 'https://example.com'
        headers = {
            'Accept': 'application/json',
            'OpenStack-API-Version': 'version header test',
        }

        session = mock.Mock()
        session.get.side_effect = [
            exceptions.Unauthorized('unauthorized'),
            # Throw a different exception the second time so that we can
            # catch it in the test and verify the retry.
            exceptions.BadRequest('bad request'),
        ]

        try:
            discover.get_version_data(
                session, url, version_header='version header test'
            )
        except exceptions.BadRequest:
            pass
        # Only one call with 'url'
        self.assertEqual(2, session.get.call_count)
        session.get.assert_has_calls(
            [
                mock.call(url, headers=headers, authenticated=None),
                mock.call(url, headers=headers, authenticated=True),
            ]
        )

    def test_endpoint_data_str(self):
        """Validate EndpointData.__str__."""
        # Populate a few fields to make sure they come through.
        epd = discover.EndpointData(
            catalog_url='abc', service_type='123', api_version=(2, 3)
        )
        exp = (
            'EndpointData{api_version=(2, 3), catalog_url=abc,'
            ' endpoint_id=None, interface=None, major_version=None,'
            ' max_microversion=None, min_microversion=None,'
            ' next_min_version=None, not_before=None, raw_endpoint=None,'
            ' region_name=None, service_id=None, service_name=None,'
            ' service_type=123, service_url=None, url=abc}'
        )
        # Works with str()
        self.assertEqual(exp, str(epd))
        # Works with implicit stringification
        self.assertEqual(exp, f"{epd}")

    def test_project_id_int_fallback(self):
        bad_url = "https://compute.example.com/v2/123456"
        epd = discover.EndpointData(catalog_url=bad_url)
        self.assertEqual((2, 0), epd.api_version)

    def test_url_version_match_project_id_int(self):
        self.session = session.Session()
        discovery_fixture = fixture.V3Discovery(V3_URL)
        discovery_doc = _create_single_version(discovery_fixture)
        self.requests_mock.get(V3_URL, status_code=200, json=discovery_doc)
        epd = discover.EndpointData(catalog_url=V3_URL).get_versioned_data(
            session=self.session, project_id='3'
        )
        self.assertEqual(epd.catalog_url, epd.url)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from keystoneauth1.access import types
from keystoneauth1 import exceptions


class ServiceProviders:
    """Helper methods for dealing with Service Providers."""

    def __init__(self, service_providers: list[types.ServiceProviderV3]):
        self._service_providers = {
            sp['id']: sp for sp in service_providers if 'id' in sp
        }

    @classmethod
    def from_token(cls, token: types.TokenResponseV3) -> 'ServiceProviders':
        if 'token' not in token:
            raise ValueError(
                'Token format does not support service providers.'
            )

        return cls(token['token'].get('service_providers', []))

    def _get_service_provider(self, sp_id: str) -> types.ServiceProviderV3:
        try:
            return self._service_providers[sp_id]
        except KeyError:
            raise exceptions.ServiceProviderNotFound(sp_id)

    def get_sp_url(self, sp_id: str) -> str:
        return self._get_service_provider(sp_id)['sp_url']

    def get_auth_url(self, sp_id: str) -> str:
        return self._get_service_provider(sp_id)['auth_url']



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing as ty

from keystoneauth1 import discover
from keystoneauth1 import plugin

if ty.TYPE_CHECKING:
    from keystoneauth1.access import access
    from keystoneauth1 import session as ks_session


class Token(plugin.BaseAuthPlugin):
    """A provider that will always use the given token and endpoint.

    This is really only useful for testing and in certain CLI cases where you
    have a known endpoint and admin token that you want to use.
    """

    def __init__(self, endpoint: str | None, token: str | None):
        super().__init__()
        # NOTE(jamielennox): endpoint is reserved for when plugins
        # can be used to provide that information
        self.endpoint = endpoint
        self.token = token

    def get_token(self, session: 'ks_session.Session') -> str | None:
        return self.token

    def get_endpoint_data(
        self,
        session: 'ks_session.Session',
        *,
        endpoint_override: str | None = None,
        discover_versions: bool = True,
        **kwargs: ty.Any,
    ) -> discover.EndpointData | None:
        """Return a valid endpoint data for a the service.

        :param session: A session object that can be used for communication.
        :type session: keystoneauth1.session.Session
        :param str endpoint_override: URL to use for version discovery other
                                      than the endpoint stored in the plugin.
                                      (optional, defaults to None)
        :param bool discover_versions: Whether to get version metadata from
                                       the version discovery document even
                                       if it major api version info can be
                                       inferred from the url.
                                       (optional, defaults to True)
        :param kwargs: Ignored.

        :raises keystoneauth1.exceptions.http.HttpError: An error from an
                                                         invalid HTTP response.

        :return: Valid EndpointData or None if not available.
        :rtype: `keystoneauth1.discover.EndpointData` or None
        """
        return super().get_endpoint_data(
            session,
            endpoint_override=endpoint_override or self.endpoint,
            discover_versions=discover_versions,
            **kwargs,
        )

    def get_endpoint(
        self, session: 'ks_session.Session', **kwargs: ty.Any
    ) -> str | None:
        """Return the supplied endpoint.

        Using this plugin the same endpoint is returned regardless of the
        parameters passed to the plugin.
        """
        return self.endpoint

    def get_auth_ref(
        self, session: 'ks_session.Session'
    ) -> ty.Optional['access.AccessInfo']:
        """Return the authentication reference of an auth plugin.

        :param session: A session object to be used for communication
        :type session: keystoneauth1.session.session
        """
        # token plugin does not have an auth ref, because it's a
        # "static" authentication using a pre-existing token.
        return None



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import queue
from threading import Thread
from timeit import default_timer as timer
from unittest import mock

import testtools

from keystoneauth1 import _fair_semaphore


class SemaphoreTests(testtools.TestCase):
    def _thread_worker(self):
        while True:
            # get returns the Item, but we don't care about the value so we
            # purposely don't assign it to anything.
            self.q.get()
            with self.s:
                self.mock_payload.do_something()
            self.q.task_done()

    # Have 5 threads do 10 different "things" coordinated by the fair
    # semaphore.
    def _concurrency_core(self, concurrency, delay):
        self.s = _fair_semaphore.FairSemaphore(concurrency, delay)

        self.q = queue.Queue()
        for i in range(5):
            t = Thread(target=self._thread_worker)
            t.daemon = True
            t.start()

        for item in range(0, 10):
            self.q.put(item)

        self.q.join()

    def setUp(self):
        super().setUp()
        self.mock_payload = mock.Mock()

    # We should be waiting at least 0.1s between operations, so
    # the 10 operations must take at *least* 1 second
    def test_semaphore_no_concurrency(self):
        start = timer()
        self._concurrency_core(None, 0.1)
        end = timer()
        self.assertTrue((end - start) > 1.0)
        self.assertEqual(self.mock_payload.do_something.call_count, 10)

    def test_semaphore_single_concurrency(self):
        start = timer()
        self._concurrency_core(1, 0.1)
        end = timer()
        self.assertTrue((end - start) > 1.0)
        self.assertEqual(self.mock_payload.do_something.call_count, 10)

    def test_semaphore_multiple_concurrency(self):
        start = timer()
        self._concurrency_core(5, 0.1)
        end = timer()
        self.assertTrue((end - start) > 1.0)
        self.assertEqual(self.mock_payload.do_something.call_count, 10)

    # do some high speed tests; I don't think we can really assert
    # much about these other than they don't deadlock...
    def test_semaphore_fast_no_concurrency(self):
        self._concurrency_core(None, 0.0)

    def test_semaphore_fast_single_concurrency(self):
        self._concurrency_core(1, 0.0)

    def test_semaphore_fast_multiple_concurrency(self):
        self._concurrency_core(5, 0.0)
